var fs = require("fs"), util = require("util"), javascriptParser = require("uglify-js").parser, astProcessor = require("uglify-js").uglify, url = require("url"), Const = require("./const"), Ast = require("./ast"), ErrCodes = require("./err_codes");

function getPageDirectiveValue(pageUrl) {
    if (!url.parse(pageUrl).protocol && !Const.RELATIVE_URL_PATTERN.test(pageUrl)) {
        return "http://" + pageUrl;
    }
    return pageUrl;
}

function getActionArgAst(arg) {
    switch (typeof arg) {
      case "number":
        return [ "num", arg ];

      case "string":
        return [ "string", arg ];

      case "object":
        var propAsts = [];
        Object.keys(arg).forEach(function(key) {
            propAsts.push([ key, getActionArgAst(arg[key]) ]);
        });
        return [ "object", propAsts ];

      default:
        return [ "name", "" + arg ];
    }
}

function createTestAst(testName, stepsInfo) {
    var stepsAst = [], stepIndex = 0;
    stepsInfo.forEach(function(info) {
        var stepFuncBodyAst = [], actionArgsAst = [], selector = info.selectors && info.selectors[info.currentSelectorIndex] && info.selectors[info.currentSelectorIndex].selector;
        if (info.elemIdentifier && selector) {
            try {
                var selectorAst = javascriptParser.parse(selector);
            } catch (parserErr) {
                throw {
                    code: ErrCodes.ELEMENT_SELECTOR_PARSING_FAILED,
                    parserErr: parserErr
                };
            }
            var selectorStatement = selectorAst[1] && selectorAst[1][0] && selectorAst[1][0][1] || [ "string", "" ];
            stepFuncBodyAst.push([ "var", [ [ info.elemIdentifier, selectorStatement ] ] ]);
            actionArgsAst.push([ "name", info.elemIdentifier ]);
        }
        if (info.nativeDialogHandlers) {
            for (var dialog in info.nativeDialogHandlers) {
                if (info.nativeDialogHandlers.hasOwnProperty(dialog)) {
                    var retValue = info.nativeDialogHandlers[dialog].retValue, handleFunctionName = Const.NATIVE_DIALOG_HANDLE_IDENTIFIER_PREFIX + dialog[0].toUpperCase() + dialog.substring(1), argumentAst = null;
                    if (typeof retValue === "string") {
                        argumentAst = [ [ "string", retValue ] ];
                    } else {
                        if (retValue === true || retValue === false || retValue === null) {
                            argumentAst = [ [ "name", retValue + "" ] ];
                        } else {
                            argumentAst = [];
                        }
                    }
                    stepFuncBodyAst.push([ "stat", [ "call", [ "name", handleFunctionName ], argumentAst ] ]);
                }
            }
        }
        if (info.actionArgs) {
            info.actionArgs.forEach(function(argNameValPair) {
                var argName = Object.keys(argNameValPair)[0];
                actionArgsAst.push(getActionArgAst(argNameValPair[argName]));
            });
        }
        stepFuncBodyAst.push([ "stat", [ "call", [ "dot", [ "name", Const.ACTIONS_OWNER_OBJECT_IDENTIFIER ], info.action ], actionArgsAst ] ]);
        var stepName = ++stepIndex + "." + info.name;
        stepsAst.push([ stepName, [ "function", null, [], stepFuncBodyAst ] ]);
    });
    return [ "stat", [ "assign", true, [ "sub", [ "string", Const.TEST_DECLARATION_MARKER ], [ "string", testName ] ], [ "object", stepsAst ] ] ];
}

function performActionOnTest(filename, testName, action, callback) {
    Ast.construct(filename, function(parseErr, ast) {
        if (parseErr) {
            callback(parseErr);
            return;
        }
        var walker = astProcessor.ast_walker(), resultAst = null, testObj = null, testNameObj = null, testIsFound = false, testNames = [];
        walker.with_walkers({
            string: function() {
                var astPath = walker.stack(), matchTestName = "";
                if (this[1] === Const.TEST_DECLARATION_MARKER) {
                    matchTestName = Ast.getAncestorByName("sub", astPath)[2];
                    testNames.push(matchTestName[1]);
                    if (matchTestName[1] === testName && Ast.isPathMatch(Const.TEST_DECLARATION_AST_PATH, walker.stack())) {
                        testIsFound = true;
                        testObj = Ast.getAncestorByName("stat", astPath);
                        testNameObj = matchTestName;
                    }
                }
            }
        }, function() {
            walker.walk(ast);
        });
        if (testIsFound) {
            if (action(testNameObj, testNames, testObj, ast)) {
                resultAst = astProcessor.gen_code(ast, {
                    beautify: true
                });
                fs.writeFile(filename, resultAst, function(writeErrs) {
                    callback(writeErrs ? {
                        code: ErrCodes.WRITE_FILE_FAILED,
                        filename: filename
                    } : null);
                });
            }
        } else {
            callback({
                code: ErrCodes.TEST_IS_NOT_FOUND,
                testName: testName
            });
        }
    });
}

exports.createFixture = function(fileName, fixtureName, fixturePage, callback) {
    var fixtureCode = util.format(Const.NEW_FIXTURE_CODE_PATTERN, fixtureName, getPageDirectiveValue(fixturePage));
    fs.writeFile(fileName, fixtureCode, function(writeErr) {
        callback(writeErr ? {
            code: ErrCodes.WRITE_FILE_FAILED,
            filename: fileName
        } : null);
    });
};

exports.editDirectives = function(fileName, directives, callback) {
    Ast.construct(fileName, function(getAstErr, ast) {
        if (getAstErr) {
            callback(getAstErr);
            return;
        }
        var walker = astProcessor.ast_walker(), fixtureDirectiveFound = false, authDirectiveIdx = -1, pageDirectiveIdx = -1;
        walker.with_walkers({
            string: function() {
                var match = Const.DIRECTIVE_EXPRESSION_PATTERN.exec(this[1]);
                if (match && Ast.isPathMatch(Const.DIRECTIVE_EXPRESSION_AST_PATH, walker.stack())) {
                    var newDirectiveValue = null, astPath = walker.stack();
                    switch (match[1]) {
                      case Const.AUTH_DIRECTIVE_LVALUE:
                        if (directives.auth) {
                            newDirectiveValue = directives.auth;
                        }
                        authDirectiveIdx = ast[1].indexOf(astPath[1]);
                        break;

                      case Const.FIXTURE_DIRECTIVE_LVALUE:
                        if (directives.fixture) {
                            newDirectiveValue = directives.fixture;
                        }
                        fixtureDirectiveFound = true;
                        break;

                      case Const.PAGE_DIRECTIVE_LVALUE:
                        if (directives.page) {
                            newDirectiveValue = getPageDirectiveValue(directives.page);
                        }
                        pageDirectiveIdx = ast[1].indexOf(astPath[1]);
                        break;

                      default:
                        return;
                    }
                    if (newDirectiveValue) {
                        this[1] = this[1].substr(0, this[1].length - match[2].length) + newDirectiveValue;
                    }
                }
            }
        }, function() {
            walker.walk(ast);
        });
        if (fixtureDirectiveFound && pageDirectiveIdx !== -1) {
            if (authDirectiveIdx > -1) {
                if (directives.auth === null) {
                    ast[1].splice(authDirectiveIdx, 1);
                }
            } else {
                if (directives.auth) {
                    var directiveAst = [ "string", util.format(Const.AUTH_DIRECTIVE_PATTERN, directives.auth) ];
                    ast[1].splice(pageDirectiveIdx + 1, 0, directiveAst);
                }
            }
            var fixtureCode = astProcessor.gen_code(ast, {
                beautify: true
            });
            fs.writeFile(fileName, fixtureCode, function(writeErr) {
                callback(writeErr ? {
                    code: ErrCodes.WRITE_FILE_FAILED,
                    filename: fileName
                } : null);
            });
        } else {
            callback({
                code: ErrCodes.INVALID_FILE_FORMAT,
                filename: fileName
            });
        }
    });
};

exports.addTest = function(fileName, testName, stepsInfo, callback) {
    Ast.construct(fileName, function(getAstErr, ast) {
        if (getAstErr) {
            callback(getAstErr);
            return;
        }
        var walker = astProcessor.ast_walker(), insertionIndex = -1, fixtureDirectiveFound = false, pageDirectiveFound = false, testNames = [];
        walker.with_walkers({
            string: function() {
                var astPath = walker.stack();
                if (this[1] === Const.TEST_DECLARATION_MARKER && Ast.isPathMatch(Const.TEST_DECLARATION_AST_PATH, astPath)) {
                    insertionIndex = ast[1].indexOf(astPath[1]);
                    testNames.push(Ast.getAncestorByName("sub", astPath)[2][1]);
                } else {
                    var match = Const.DIRECTIVE_EXPRESSION_PATTERN.exec(this[1]);
                    if (match && Ast.isPathMatch(Const.DIRECTIVE_EXPRESSION_AST_PATH, astPath) && Const.DIRECTIVE_LVALUES.indexOf(match[1]) > -1) {
                        if (match[1] === Const.FIXTURE_DIRECTIVE_LVALUE) {
                            fixtureDirectiveFound = true;
                        } else {
                            if (match[1] === Const.PAGE_DIRECTIVE_LVALUE) {
                                pageDirectiveFound = true;
                            }
                        }
                        insertionIndex = ast[1].indexOf(astPath[1]);
                    }
                }
            }
        }, function() {
            walker.walk(ast);
        });
        if (insertionIndex === -1 || !fixtureDirectiveFound || !pageDirectiveFound) {
            callback({
                code: ErrCodes.INVALID_FILE_FORMAT,
                filename: fileName
            });
            return;
        }
        if (testNames.indexOf(testName) > -1) {
            callback({
                code: ErrCodes.TEST_NAME_CHANGED_TO_ALREADY_USED,
                testName: testName
            });
            return;
        }
        try {
            var testAst = createTestAst(testName, stepsInfo);
        } catch (testAstCreationErr) {
            callback(testAstCreationErr);
            return;
        }
        ast[1].splice(insertionIndex + 1, 0, testAst);
        var fixtureCode = astProcessor.gen_code(ast, {
            beautify: true
        });
        fs.writeFile(fileName, fixtureCode, function(writeErr) {
            callback(writeErr ? {
                code: ErrCodes.WRITE_FILE_FAILED,
                filename: fileName
            } : null);
        });
    });
};

exports.deleteTest = function(filename, testName, callback) {
    var action = function(match, testNames, stack, ast) {
        var testIndex = ast[1].indexOf(stack);
        delete ast[1][testIndex];
        return true;
    };
    performActionOnTest(filename, testName, action, function(res) {
        callback(res);
    });
};

exports.renameTest = function(filename, oldName, newName, callback) {
    var action = function(match, testNames) {
        if (testNames.indexOf(newName) === -1) {
            match[1] = newName;
            return true;
        } else {
            callback({
                code: ErrCodes.TEST_NAME_CHANGED_TO_ALREADY_USED,
                line: match[0].start.line,
                testName: newName
            });
            return false;
        }
    };
    performActionOnTest(filename, oldName, action, function(res) {
        callback(res);
    });
};