var fs = require("fs"), path = require("path"), javascriptParser = require("uglify-js").parser, astProcessor = require("uglify-js").uglify, async = require("async"), Const = require("./const"), Ast = require("./ast"), FuncCallValidator = require("./func_call_validator"), ErrCodes = require("./err_codes");

var Compiler = module.exports = function(filename, modules, requiresCodeCache) {
    requiresCodeCache = requiresCodeCache || {};
    this.walker = astProcessor.ast_walker();
    this.filename = filename;
    this.workingDir = path.dirname(this.filename);
    this.modules = modules;
    this.requiresCodeCache = requiresCodeCache;
    this.requires = [];
    this.hasReqErrs = false;
    this.line = 0;
    this.errs = [];
    this.out = {
        fixture: "",
        page: "",
        authCredentials: null,
        sharedJs: "",
        testsStepData: {}
    };
};

Compiler.prototype._addRequire = function(require) {
    if (this.requires.indexOf(require) > -1) {
        this.errs.push({
            code: ErrCodes.REQUIRED_FILE_ALREADY_INCLUDED,
            filename: this.filename,
            line: this.line,
            req: this.require
        });
    } else {
        this.requires.push(require);
    }
};

Compiler.prototype._compileDirective = function(match) {
    switch (match[1]) {
      case Const.AUTH_DIRECTIVE_LVALUE:
        if (this.out.authCredentials) {
            this.errs.push({
                code: ErrCodes.AUTH_DIRECTIVE_REDEFINITION,
                filename: this.filename,
                line: this.line
            });
            break;
        }
        var credentials = match[2].split(Const.AUTH_CREDENTIALS_SEPARATOR);
        if (credentials.length < 2) {
            this.errs.push({
                code: ErrCodes.INVALID_NETWORK_AUTHENTICATION_CREDENTIALS_FORMAT,
                filename: this.filename,
                line: this.line
            });
            break;
        }
        this.out.authCredentials = {
            username: credentials[0],
            password: credentials[1]
        };
        break;

      case Const.FIXTURE_DIRECTIVE_LVALUE:
        if (this.out.fixture) {
            this.errs.push({
                code: ErrCodes.FIXTURE_DIRECTIVE_REDEFINITION,
                filename: this.filename,
                line: this.line
            });
            break;
        }
        this.out.fixture = match[2];
        break;

      case Const.PAGE_DIRECTIVE_LVALUE:
        if (this.out.page) {
            this.errs.push({
                code: ErrCodes.PAGE_DIRECTIVE_REDEFINITION,
                filename: this.filename,
                line: this.line
            });
            break;
        }
        this.out.page = match[2];
        break;

      case Const.REQUIRE_DIRECTIVE_LVALUE:
        if (match[2].indexOf(Const.MODULE_PREFIX) === 0) {
            var moduleName = match[2].slice(1), moduleFiles = this.modules && this.modules[moduleName];
            if (!moduleFiles) {
                this.errs.push({
                    code: ErrCodes.MODULE_NOT_FOUND,
                    filename: this.filename,
                    line: this.line,
                    moduleName: moduleName
                });
                break;
            }
            var compiler = this;
            moduleFiles.forEach(function(moduleFile) {
                compiler._addRequire(moduleFile);
            });
            break;
        }
        var require = path.join(this.workingDir, match[2]);
        this._addRequire(require);
        break;

      default:
        return false;
    }
    if (!Ast.isPathMatch(Const.DIRECTIVE_EXPRESSION_AST_PATH, this.walker.stack())) {
        this.errs.push({
            code: ErrCodes.MISPLACED_DIRECTIVE,
            filename: this.filename,
            line: this.line
        });
    }
    return true;
};

Compiler.prototype._compileTest = function() {
    var astPath = this.walker.stack();
    if (!Ast.isPathMatch(Const.TEST_DECLARATION_AST_PATH, astPath)) {
        this.errs.push({
            code: ErrCodes.MISPLACED_TEST_DECLARATION,
            filename: this.filename,
            line: this.line
        });
        return;
    }
    var testName = Ast.getAncestorByName("sub", astPath)[2][1];
    if (!testName) {
        this.errs.push({
            code: ErrCodes.EMPTY_TEST_NAME,
            filename: this.filename,
            line: this.line
        });
        return;
    }
    if (this.out.testsStepData[testName]) {
        this.errs.push({
            code: ErrCodes.DUPLICATE_TEST_NAME,
            filename: this.filename,
            testName: testName,
            line: this.line
        });
        return;
    }
    var obj = Ast.getAncestorByName("assign", astPath)[3];
    if (Ast.getEntryName(obj) !== "object") {
        this.errs.push({
            code: ErrCodes.INVALID_TEST_ASSIGNMENT,
            filename: this.filename,
            line: this.line
        });
        return;
    }
    this._compileTestSteps(testName, obj[1]);
};

Compiler.prototype._compileTestSteps = function(testName, steps) {
    if (!steps.length) {
        this.errs.push({
            code: ErrCodes.TEST_IS_EMPTY,
            filename: this.filename,
            line: this.line
        });
        return;
    }
    var stepAsts = [], compiler = this;
    this.out.testsStepData[testName] = {
        names: [],
        js: null
    };
    steps.forEach(function(step) {
        var stepAst = step[1];
        if (Ast.getEntryName(stepAst) !== "function") {
            compiler.errs.push({
                code: ErrCodes.TEST_STEP_IS_NOT_A_FUNCTION,
                filename: compiler.filename,
                line: compiler.line
            });
            return;
        }
        var funcBody = stepAst[3], funcLastEntr = funcBody[funcBody.length - 1];
        if (!funcLastEntr) {
            compiler.errs.push({
                code: ErrCodes.TEST_STEP_IS_EMPTY,
                filename: compiler.filename,
                line: compiler.line
            });
            return;
        }
        funcLastEntr[0].last = true;
        FuncCallValidator.validate(stepAst, compiler.filename, compiler.errs, false);
        stepAsts.push(stepAst);
        compiler.out.testsStepData[testName].names.push(step[0]);
    });
    if (!this.errs.length) {
        this.out.testsStepData[testName].js = astProcessor.gen_code([ "array", stepAsts ], {
            beautify: true
        });
    }
};

Compiler.prototype._getRemainderCode = function(ast) {
    var astBranches = [];
    ast[1].forEach(function(statement) {
        if (!statement[0].remove) {
            astBranches.push(statement);
        }
    });
    if (astBranches.length) {
        var remainderAst = [ "toplevel", astBranches ];
        FuncCallValidator.validate(remainderAst, this.filename, this.errs, true);
        return this.errs.length ? "" : astProcessor.gen_code(remainderAst, {
            beautify: true
        });
    }
    return "";
};

Compiler.prototype._readAndValidateRequireFile = function(require, callback) {
    var compiler = this;
    fs.readFile(require, function(readErr, data) {
        var reqErrs = [];
        if (readErr) {
            reqErrs.push({
                code: ErrCodes.FAILED_LOAD_REQUIRE,
                filename: compiler.filename,
                req: require
            });
        } else {
            try {
                var jsCode = data.toString().trim(), requireAst = javascriptParser.parse(jsCode, false, true);
            } catch (parserErr) {
                reqErrs.push({
                    code: ErrCodes.JAVASCRIPT_PARSING_FAILED,
                    filename: require,
                    parserErr: parserErr
                });
            }
            if (requireAst) {
                FuncCallValidator.validate(requireAst, require, reqErrs, true);
            }
        }
        compiler.errs = compiler.errs.concat(reqErrs);
        if (jsCode && jsCode[jsCode.length - 1] !== ";") {
            jsCode += ";";
        }
        compiler.requiresCodeCache[require] = {
            hasErrs: !!reqErrs.length,
            jsCode: jsCode
        };
        callback(null, compiler.requiresCodeCache[require].hasErrs ? null : jsCode);
    });
};

Compiler.prototype._getRequiresCode = function(callback) {
    var compiler = this, readers = [];
    this.requires.forEach(function(require) {
        readers.push(function(readerCallback) {
            if (compiler.requiresCodeCache[require]) {
                var cache = compiler.requiresCodeCache[require];
                compiler.hasReqErrs |= cache.hasErrs;
                readerCallback(null, cache.hasErrs ? null : cache.jsCode);
            } else {
                compiler._readAndValidateRequireFile(require, readerCallback);
            }
        });
    });
    async.parallel(readers, function(asyncErr, requiresJs) {
        callback(compiler.errs.length ? "" : requiresJs.join(""));
    });
};

Compiler.prototype._analyzeAst = function(ast) {
    var compiler = this;
    this.walker.with_walkers({
        string: function() {
            var astPath = compiler.walker.stack(), topStatement = astPath[1][0];
            compiler.line = Ast.getCurrentSrcLineNum(astPath);
            if (this[1] === Const.TEST_DECLARATION_MARKER) {
                topStatement.remove = true;
                compiler._compileTest();
                return;
            }
            var match = Const.DIRECTIVE_EXPRESSION_PATTERN.exec(this[1]);
            if (match) {
                topStatement.remove = compiler._compileDirective(match);
            }
        }
    }, function() {
        compiler.walker.walk(ast);
    });
};

Compiler.prototype.compile = function(callback) {
    var compiler = this;
    Ast.construct(this.filename, function(parseErr, ast) {
        if (parseErr) {
            callback([ parseErr ]);
        } else {
            compiler._analyzeAst(ast);
            if (!compiler.out.fixture) {
                compiler.errs.push({
                    code: ErrCodes.FIXTURE_DIRECTIVE_IS_UNDEFINED,
                    filename: compiler.filename
                });
            }
            if (!compiler.out.page) {
                compiler.errs.push({
                    code: ErrCodes.PAGE_DIRECTIVE_IS_UNDEFINED,
                    filename: compiler.filename
                });
            }
            var remainderJs = compiler._getRemainderCode(ast);
            compiler._getRequiresCode(function(requireJs) {
                if (!compiler.errs.length && !compiler.hasReqErrs) {
                    compiler.out.sharedJs = [ requireJs, remainderJs ].join("");
                    callback(null, compiler.out);
                } else {
                    callback(compiler.errs);
                }
            });
        }
    });
};