var url = require("url"), util = require("util"), errMsgBuilder = require("./err_msg_builder"), moment = require("moment"), EventEmitter = require("events").EventEmitter;

function hasTaskErrors(taskReport) {
    return Object.keys(taskReport.testErrReports).length > 0;
}

var Reporter = exports.Reporter = function() {
    EventEmitter.call(this);
    this.taskReports = [];
};

util.inherits(Reporter, EventEmitter);

Reporter.prototype.getTaskReportByUid = function(taskUid) {
    for (var i = 0; i < this.taskReports.length; i++) {
        if (this.taskReports[i].uid === taskUid) {
            return this.taskReports[i];
        }
    }
};

Reporter.prototype.getTaskReports = function() {
    return this.taskReports;
};

Reporter.prototype.taskAdded = function(taskUid, name, testCount, workerNames, reportUnstable) {
    var taskReport = {
        uid: taskUid,
        name: name,
        status: "pending",
        startedAt: null,
        completedAt: null,
        workerNames: workerNames,
        testCount: testCount,
        failed: 0,
        passed: 0,
        testErrReports: {}
    };
    if (reportUnstable) {
        taskReport.unstableTests = {};
    }
    this.completedRunsPerTest = {};
    this.taskReports.push(taskReport);
    this.emit("taskUpdated", {
        id: taskUid
    });
};

Reporter.prototype.taskStarted = function(taskUid) {
    var taskReport = this.getTaskReportByUid(taskUid);
    taskReport.startedAt = moment().format("L LT");
    if (taskReport.startedAt) {
        taskReport.startedAt = taskReport.startedAt.split(" ");
    }
    taskReport.status = "running";
    this.emit("taskUpdated", {
        id: taskUid
    });
};

Reporter.prototype.taskCompleted = function(taskUid) {
    var taskReport = this.getTaskReportByUid(taskUid);
    taskReport.completedAt = taskReport.startedAt && moment().format("L LT");
    if (taskReport.completedAt) {
        taskReport.completedAt = taskReport.completedAt.split(" ");
    }
    taskReport.status = hasTaskErrors(taskReport) ? "failed" : "succeeded";
    this.emit("taskUpdated", {
        id: taskUid
    });
};

Reporter.prototype.testRunCompleted = function(testRun) {
    var taskReport = this.getTaskReportByUid(testRun.taskUid), testUid = testRun.test.uid;
    if (testRun.errs.length) {
        this._processTestRunErrs(testRun, taskReport, testUid);
        this.emit("taskDetailErrUpdated", {
            id: testRun.taskUid
        });
    }
    if (taskReport.unstableTests && testRun.quarantine.failed && testRun.quarantine.succeeded) {
        if (!taskReport.unstableTests[testUid]) {
            taskReport.unstableTests[testUid] = {
                name: testRun.test.name,
                fixtureName: testRun.fixture.name,
                fixturePath: testRun.fixture.path.join("/"),
                workerNames: []
            };
        }
        taskReport.unstableTests[testUid].workerNames.push(testRun.workerName);
    }
    if (taskReport.testErrReports[testUid]) {
        return;
    }
    if (!this.completedRunsPerTest[testRun.taskUid]) {
        this.completedRunsPerTest[testRun.taskUid] = {};
    }
    this.completedRunsPerTest[testRun.taskUid][testUid] = this.completedRunsPerTest[testRun.taskUid][testUid] || 0;
    if (++this.completedRunsPerTest[testRun.taskUid][testUid] === taskReport.workerNames.length) {
        taskReport.passed++;
        if (taskReport.passed + taskReport.failed < taskReport.testCount) {
            this.emit("taskUpdated", {
                id: testRun.taskUid
            });
        }
    }
};

Reporter.prototype._processTestRunErrs = function(testRun, taskReport, testUid) {
    if (!taskReport.testErrReports[testUid]) {
        taskReport.testErrReports[testUid] = {
            name: testRun.test.name,
            fixtureName: testRun.fixture.name,
            fixturePath: testRun.fixture.path.join("/"),
            errs: []
        };
        taskReport.failed++;
        this.emit("taskListTestFailed", {
            id: testRun.taskUid
        });
    }
    testRun.errs.forEach(function(err) {
        var reportItem = {
            msg: errMsgBuilder.build(err.err),
            workerName: testRun.workerName,
            possibleCauses: []
        };
        err.possibleCauses.forEach(function(cause) {
            reportItem.possibleCauses.push({
                type: cause.type,
                msg: errMsgBuilder.build(cause.err)
            });
        });
        taskReport.testErrReports[testUid].errs.push(reportItem);
    });
};