var EventEmitter = require("events").EventEmitter, exec = require("child_process").exec, fs = require("fs"), moment = require("moment"), os = require("os"), path = require("path"), util = require("util"), Api = require("./api").Api, AssetsManager = require("./assets_manager"), Config = require("./../config").Config, ControlPanel = require("./control_panel").ControlPanel, CookieShelf = require("./cookie_shelf").CookieShelf, errMsgBuilder = require("./err_msg_builder"), Proxy = require("../proxy/"), Recorder = require("./recorder").Recorder, Reporter = require("./reporter").Reporter, TestRunner = require("./test_runner").TestRunner, VirtualFS = require("./vfs").VirtualFS, WorkerPool = require("./worker_pool").WorkerPool;

var GREETING = "DevExpress TestCafe - functional testing for the web.\n" + "Copyright (C) 2012 - %s Developer Express Inc.\n" + "-----------------------------------------------\n", CONTROL_PANEL_URL_INFO_PATTERN = "Control Panel URL - http://%s\n\n", SHUTDOWN_MSG_PATTERN = "%s - Initialization Error:\n%s\n";

var TestCafe = exports.TestCafe = function(startupCfg, standalone) {
    EventEmitter.call(this);
    this.controlPanel = null;
    this.proxy = null;
    this.testRunner = null;
    this.vfs = null;
    this.workerPool = null;
    this.assetsManager = null;
    this._init(startupCfg, standalone);
};

util.inherits(TestCafe, Api);

TestCafe.prototype._fatalError = function(errMsg, standalone) {
    var shutDownMsg = util.format(SHUTDOWN_MSG_PATTERN, moment().format("L LT"), errMsg);
    this.close();
    if (standalone) {
        util.error(shutDownMsg);
        process.exit(1);
    } else {
        throw shutDownMsg;
    }
};

TestCafe.prototype.close = function() {
    if (this.vfs) {
        this.vfs.removeFSWatchers();
    }
    if (this.workerPool) {
        this.workerPool.removeAllWorkers();
    }
    if (this.proxy) {
        try {
            this.proxy.close();
        } catch (err) {}
    }
    if (this.controlPanel) {
        try {
            this.controlPanel.close();
        } catch (err) {}
    }
};

TestCafe.prototype._init = function(startupCfg, standalone) {
    var testCafe = this;
    if (standalone) {
        util.print(util.format(GREETING, new Date().getFullYear()));
    }
    testCafe.config = new Config();
    testCafe._initConfig(startupCfg, function(config, err) {
        if (err) {
            var messages = errMsgBuilder.build(util.isArray(err) ? err : [ err ]);
            testCafe._fatalError(messages.join("\n"), standalone);
        }
        var reporter = new Reporter(), cookieShelf = new CookieShelf();
        testCafe.proxy = new Proxy(config.servicePort, config.hostname, cookieShelf);
        testCafe.proxy.events.broadcast.listen("fatalError", function(err) {
            testCafe._fatalError(errMsgBuilder.build(err), standalone);
        });
        testCafe.assetsManager = new AssetsManager(testCafe.proxy.serviceChannel);
        try {
            testCafe.vfs = new VirtualFS(config.testsDir, testCafe.config);
        } catch (err) {
            testCafe._fatalError(errMsgBuilder.build(err), standalone);
        }
        testCafe.workerPool = new WorkerPool(config);
        testCafe.testRunner = new TestRunner(testCafe.proxy, reporter, testCafe.workerPool, cookieShelf, testCafe.assetsManager, config.controlPanelHost);
        var recorder = new Recorder(testCafe.proxy, testCafe.vfs, cookieShelf, testCafe.assetsManager);
        Api.call(testCafe, testCafe.testRunner, testCafe.workerPool, reporter, testCafe.vfs, config, standalone);
        testCafe.controlPanel = new ControlPanel(config.controlPanelPort, testCafe.testRunner, testCafe.vfs, testCafe.workerPool, reporter, testCafe, recorder);
        testCafe.controlPanel.on("fatalError", function(err) {
            testCafe._fatalError(errMsgBuilder.build(err), standalone);
        });
        testCafe.proxy.start();
        testCafe.controlPanel.start();
        if (standalone) {
            var controlPanelHost = config.hostname === "127.0.0.1" ? "localhost:" + config.controlPanelPort : config.controlPanelHost;
            util.print(util.format(CONTROL_PANEL_URL_INFO_PATTERN, controlPanelHost));
            var controlPanelUrl = "http://" + controlPanelHost.replace(/"/, '\\"'), platform = os.platform();
            if (platform.match(/^win/)) {
                exec('start "" "' + controlPanelUrl + '"');
            } else {
                if (platform === "darwin") {
                    exec("open " + controlPanelUrl);
                } else {
                    exec("xdg-open " + controlPanelUrl);
                }
            }
        }
    });
};

TestCafe.prototype._initConfig = function(startupCfg, callback) {
    if (startupCfg) {
        var result = this.config.initFromObj(startupCfg);
        callback(result.cfg, result.errs);
    } else {
        this.config.initFromFile(callback);
    }
};