var EventEmitter = require("events").EventEmitter, path = require("path"), url = require("url"), util = require("util"), uuid = require("node-uuid"), CLIENT_ERR = require("../shared/client_errs"), cmd = require("../shared/service_msg_cmd"), ERR = require("./server_errs"), injector = require("./injector");

var PAGE_ERROR_VERIFICATION_HTML_TEMPLATE = "/*" + "<!DOCTYPE html>" + "<html>" + "<head>" + "<title></title>" + '<script type="text/javascript">%s</script>' + "</head>" + "<body></body>" + "</html>" + "*/", QUARANTINE_TEST_RUN_COUNT = 3;

function proxyErrRes(proxyCtx, resBody) {
    proxyCtx.res.statusCode = 500;
    if (resBody) {
        proxyCtx.res.setHeader("content-type", "text/html");
        proxyCtx.res.end(resBody);
    } else {
        proxyCtx.res.end();
    }
}

var TestRunner = exports.TestRunner = function(proxy, reporter, workerPool, cookieShelf, assetsManager, controlPanelHost) {
    EventEmitter.call(this);
    this.controlPanelHost = controlPanelHost;
    this.reporter = reporter;
    this.proxy = proxy;
    this.workerPool = workerPool;
    this.cookieShelf = cookieShelf;
    this.assetsManager = assetsManager;
    this.tasks = {};
    this.testRuns = {};
    this.workerPool.on("workerDisconnected", this._onWorkerDisconnected.bind(this));
    var proxyEvents = this.proxy.events.for(TestRunner.JOB_OWNER_TOKEN);
    proxyEvents.listen("authCredentialsRequested", this._onAuthCredentialsRequested.bind(this));
    proxyEvents.listen("beforeProxyResponse", this._onBeforeProxyResponse.bind(this));
    proxyEvents.listen("error", this._onProxyError.bind(this));
    var serviceChannelEvents = this.proxy.serviceChannel.events.for(TestRunner.JOB_OWNER_TOKEN);
    serviceChannelEvents.listen("serviceMsg", this._onServiceMsg.bind(this));
    serviceChannelEvents.listen("testCafeScriptRequested", this._onTestCafeScriptRequested.bind(this));
};

util.inherits(TestRunner, EventEmitter);

TestRunner.JOB_OWNER_TOKEN = "p";

TestRunner.MAX_TEST_RUN_RESTARTS = 3;

TestRunner.POSSIBLE_FAIL_CAUSE_TYPES = {
    JS_ERROR: "JS_ERROR",
    FAILED_RESOURCE_REQUEST: "FAILED_RESOURCE_REQUEST"
};

TestRunner.prototype._onAuthCredentialsRequested = function(proxyCtx, callback) {
    var testRun = this.testRuns[proxyCtx.jobInfo.uid];
    callback(testRun ? testRun.authCredentials : null);
};

TestRunner.prototype._onBeforeProxyResponse = function(ctx, callback) {
    var testRun = this.testRuns[ctx.jobInfo.uid];
    if (!testRun) {
        proxyErrRes(ctx);
        return;
    }
    if (ctx.originRes.statusCode >= 400) {
        var err = {
            code: ERR.TEST_RUNNER_RESOURCE_ERROR_HTTP_STATUS_CODE,
            url: ctx.originResourceInfo.url,
            statusCode: ctx.originRes.statusCode
        };
        if (ctx.originResContentInfo.isPage && !ctx.isIFrame) {
            this._addTestError(testRun, err);
            testRun.halted = true;
            this.workerPool.redirectToWorkerIdle(ctx, testRun.workerName);
            return;
        } else {
            this._addPossibleTestFailCause(testRun, err);
        }
    }
    if (ctx.originResContentInfo.isPage && ctx.originRes.statusCode === 200) {
        if (testRun.halted) {
            this.workerPool.redirectToWorkerIdle(ctx, testRun.workerName);
        } else {
            this._injectTest(ctx, testRun, callback);
        }
    } else {
        callback();
    }
};

TestRunner.prototype._onProxyError = function(proxyCtx, err) {
    var testRun = this.testRuns[proxyCtx.jobInfo.uid];
    if (testRun) {
        if (proxyCtx.originResContentInfo && proxyCtx.originResContentInfo.isPage) {
            this._addTestError(testRun, err);
            testRun.halted = true;
            this.workerPool.redirectToWorkerIdle(proxyCtx, testRun.workerName);
        } else {
            if (proxyCtx.isPageReqCandidate) {
                var causeUid = this._addPossibleTestFailCause(testRun, err);
                var errHandlerScript = this.assetsManager.compileTemplate(this.assetsManager.TEMPLATES.TEST_RUN_ERR_HANDLER, {
                    workerIdleUrl: this.workerPool.getWorkerIdleUrl(testRun.workerName),
                    jobUid: testRun.uid,
                    jobOwnerToken: TestRunner.JOB_OWNER_TOKEN,
                    serviceMsgUrl: this.proxy.serviceChannel.serviceMsgUrl,
                    error: JSON.stringify(err),
                    removeFailCauseUid: causeUid
                });
                proxyErrRes(proxyCtx, util.format(PAGE_ERROR_VERIFICATION_HTML_TEMPLATE, errHandlerScript));
            } else {
                this._addPossibleTestFailCause(testRun, err);
                proxyErrRes(proxyCtx);
            }
        }
    }
};

TestRunner.prototype._onTestCafeScriptRequested = function(refererInfo, callback) {
    var runner = this, testRun = this.testRuns[refererInfo.jobInfo.uid];
    this.cookieShelf.getClientCookieString(refererInfo.jobInfo, refererInfo.originResourceInfo.url, function(cookie) {
        var testRunScript = runner.assetsManager.compileTemplate(runner.assetsManager.TEMPLATES.TEST_RUN, {
            stepNames: refererInfo.isIFrame ? "null" : JSON.stringify(testRun.test.stepData.names),
            testSteps: refererInfo.isIFrame ? "null" : testRun.test.stepData.js,
            sharedJs: refererInfo.isIFrame ? "null" : testRun.fixture.sharedJs,
            nextStep: testRun.nextStep,
            workerIdleUrl: runner.workerPool.getWorkerIdleUrl(testRun.workerName),
            restartTestRunUrl: runner.workerPool.getRestartTestRunUrl(testRun.workerName),
            halted: testRun.halted,
            jobUid: testRun.uid,
            jobOwnerToken: TestRunner.JOB_OWNER_TOKEN,
            serviceMsgUrl: runner.proxy.serviceChannel.serviceMsgUrl,
            cookie: cookie.replace(/'/g, "\\'"),
            originHost: refererInfo.originResourceInfo.host,
            originProtocol: refererInfo.originResourceInfo.protocol,
            originHostname: refererInfo.originResourceInfo.hostname,
            originPort: refererInfo.originResourceInfo.port || "",
            emulateCursor: testRun.emulateCursor
        });
        callback(testRunScript);
    });
};

TestRunner.prototype._onServiceMsg = function(msg, callback) {
    var res = null, expectedInactivityDuration = null, testRun = this.testRuns[msg.jobUid];
    if (testRun && !testRun.halted) {
        switch (msg.cmd) {
          case cmd.TEST_FAIL:
            this._addTestError(testRun, msg.err);
            testRun.halted = true;
            break;

          case cmd.ASSERTION_FAILED:
            this._addTestError(testRun, msg.err);
            break;

          case cmd.POSSIBLE_FAIL_CAUSE_FOUND:
            this._addPossibleTestFailCause(testRun, msg.err);
            break;

          case cmd.REMOVE_POSSIBLE_FAIL_CAUSE:
            this._removePossibleTestFailCause(testRun, msg.causeUid);
            break;

          case cmd.TEST_COMPLETE:
            testRun.halted = true;
            break;

          case cmd.SET_STEPS_SHARED_DATA:
            testRun.stepsSharedData = msg.stepsSharedData;
            break;

          case cmd.GET_STEPS_SHARED_DATA:
            res = testRun.stepsSharedData;
            break;

          case cmd.SET_NEXT_STEP:
            testRun.nextStep = msg.nextStep;
            break;

          case cmd.INACTIVITY_EXPECTED:
            expectedInactivityDuration = msg.duration;
            break;

          default:
            break;
        }
        this.workerPool.workerHeartbeat(testRun.workerName, expectedInactivityDuration);
    }
    callback(res);
};

TestRunner.prototype._addPossibleTestFailCause = function(testRun, err) {
    var uid = uuid.v4(), type = err.code === CLIENT_ERR.UNCAUGHT_JS_ERROR ? TestRunner.POSSIBLE_FAIL_CAUSE_TYPES.JS_ERROR : TestRunner.POSSIBLE_FAIL_CAUSE_TYPES.FAILED_RESOURCE_REQUEST;
    testRun.possibleFailCauses.push({
        uid: uid,
        type: type,
        err: err
    });
    return uid;
};

TestRunner.prototype._removePossibleTestFailCause = function(testRun, causeUid) {
    var idx = -1;
    for (var i = 0; i < testRun.possibleFailCauses.length; i++) {
        if (testRun.possibleFailCauses[i].uid === causeUid) {
            idx = i;
            break;
        }
    }
    if (idx > -1) {
        testRun.possibleFailCauses.splice(idx, 1);
    }
};

TestRunner.prototype._addTestError = function(testRun, err) {
    testRun.errs.push({
        err: err,
        possibleCauses: testRun.possibleFailCauses
    });
};

TestRunner.prototype._injectTest = function(proxyCtx, testRun, callback) {
    var runner = this;
    if (!proxyCtx.isIFrame) {
        testRun.possibleFailCauses = [];
    }
    var opt = {
        urlReplacer: this.proxy.getResourceUrlReplacer(proxyCtx),
        styleUrl: this.assetsManager.uiStyleUrl,
        scriptUrl: this.proxy.serviceChannel.testCafeScriptUrl
    };
    injector.injectInPage(proxyCtx.originResBodyChunks, proxyCtx.originResContentInfo.encoding, proxyCtx.originResContentInfo.charset, opt, function(err, html) {
        if (err) {
            testRun.halted = true;
            runner._addTestError(testRun, err);
            runner.workerPool.redirectToWorkerIdle(proxyCtx, testRun.workerName);
        } else {
            proxyCtx.originResBodyChunks = [ html ];
            callback();
        }
    });
};

TestRunner.prototype._getNextTestRunUrl = function(worker) {
    var testRunUid = worker.testRuns.length && worker.testRuns[0];
    if (testRunUid) {
        var testRun = this.testRuns[testRunUid], task = this.tasks[testRun.taskUid];
        if (!task.started) {
            this.reporter.taskStarted(testRun.taskUid);
        }
        task.started = true;
        worker.currentTest = {
            name: testRun.test.name,
            filename: testRun.fixture.fileName,
            path: testRun.fixture.path
        };
        try {
            return this.proxy.getProxyUrl(testRun.fixture.page, testRunUid, TestRunner.JOB_OWNER_TOKEN);
        } catch (err) {
            this._addTestError(testRun, err);
            this._completeTestRun(testRun);
        }
    }
    return "";
};

TestRunner.prototype.getWorkerStatus = function(workerName) {
    var worker = this.workerPool.get(workerName);
    if (!worker) {
        return;
    }
    this._completeHaltedTestRun(worker);
    var testRunUrl = this._getNextTestRunUrl(worker);
    return this.workerPool.getWorkerStatus(workerName, encodeURI(testRunUrl));
};

TestRunner.prototype.addTask = function(name, suite, testUids, workerNames, emulateCursor, quarantineMode) {
    var runner = this, taskUid = uuid.v4(), task = {
        pendingTestRunCount: 0,
        started: false,
        quarantineMode: quarantineMode
    };
    workerNames.forEach(function(workerName) {
        var worker = runner.workerPool.get(workerName);
        testUids.forEach(function(testUid) {
            var testRunUid = uuid.v4(), test = suite.tests[testUid], fixture = suite.fixtures[test.fixtureUid];
            runner.testRuns[testRunUid] = {
                authCredentials: fixture.authCredentials,
                emulateCursor: emulateCursor,
                errs: [],
                fixture: fixture,
                halted: false,
                nextStep: 0,
                possibleFailCauses: [],
                restartCount: 0,
                stepsSharedData: null,
                taskUid: taskUid,
                test: test,
                uid: testRunUid,
                workerName: workerName,
                quarantine: {
                    succeeded: 0,
                    failed: 0,
                    lastFailedRunErrs: []
                }
            };
            worker.testRuns.push(testRunUid);
            task.pendingTestRunCount++;
        });
    });
    this.tasks[taskUid] = task;
    this.reporter.taskAdded(taskUid, name, testUids.length, workerNames, quarantineMode);
    var taskReport = runner.reporter.getTaskReportByUid(taskUid);
    workerNames.forEach(function(workerName) {
        var worker = runner.workerPool.get(workerName);
        worker.taskReport = taskReport;
    });
    this.emit("taskAdded", workerNames);
    return taskUid;
};

TestRunner.prototype._completeHaltedTestRun = function(worker) {
    var testRunUid = worker.testRuns.length && worker.testRuns[0];
    if (testRunUid) {
        var testRun = this.testRuns[testRunUid];
        if (testRun.halted) {
            this._completeTestRun(testRun);
        }
    }
};

TestRunner.prototype._quarantine = function(testRun) {
    if (testRun.errs.length) {
        testRun.quarantine.lastFailedRunErrs = testRun.errs;
        testRun.quarantine.failed++;
    } else {
        testRun.quarantine.succeeded++;
    }
    if (testRun.quarantine.failed > 0) {
        if (testRun.quarantine.succeeded + testRun.quarantine.failed < QUARANTINE_TEST_RUN_COUNT) {
            this._resetTestRun(testRun);
            return true;
        } else {
            testRun.errs = testRun.quarantine.failed > testRun.quarantine.succeeded ? testRun.quarantine.lastFailedRunErrs : [];
        }
    }
    return false;
};

TestRunner.prototype._completeTestRun = function(testRun) {
    var task = this.tasks[testRun.taskUid], workerName = testRun.workerName, worker = this.workerPool.get(workerName);
    if (task.quarantineMode && this._quarantine(testRun)) {
        return;
    }
    this.cookieShelf.removeCookies({
        ownerToken: TestRunner.JOB_OWNER_TOKEN,
        uid: testRun.uid
    });
    this.reporter.testRunCompleted(testRun);
    worker.testRuns.splice(worker.testRuns.indexOf(testRun.uid), 1);
    if (!worker.testRuns.length) {
        this.workerPool.onWorkerTaskComplete(worker);
    }
    delete this.testRuns[testRun.uid];
    task.pendingTestRunCount--;
    if (!task.pendingTestRunCount) {
        this.reporter.taskCompleted(testRun.taskUid);
        delete this.tasks[testRun.taskUid];
    } else {
        worker.taskReport = this.reporter.getTaskReportByUid(testRun.taskUid);
    }
};

TestRunner.prototype._onWorkerDisconnected = function(err, worker) {
    var runner = this, runsToComplete = worker.testRuns.slice(0);
    runsToComplete.forEach(function(testRunUid) {
        var testRun = runner.testRuns[testRunUid];
        runner._addTestError(testRun, err);
        runner._completeTestRun(testRun);
    });
};

TestRunner.prototype._resetTestRun = function(testRun) {
    testRun.errs = [];
    testRun.halted = false;
    testRun.nextStep = 0;
    testRun.stepsSharedData = null;
    this.cookieShelf.removeCookies({
        ownerToken: TestRunner.JOB_OWNER_TOKEN,
        uid: testRun.uid
    });
};

TestRunner.prototype.restartCurrentTestRun = function(workerName) {
    var worker = this.workerPool.get(workerName), testRunUid = worker.testRuns.length && worker.testRuns[0], testRun = this.testRuns[testRunUid];
    if (testRun) {
        if (++testRun.restartCount > TestRunner.MAX_TEST_RUN_RESTARTS) {
            this._addTestError(testRun, {
                code: ERR.TEST_RUNNER_MAX_TEST_RUN_RESTARTS
            });
            testRun.halted = true;
        } else {
            this._resetTestRun(testRun);
        }
    }
};